function [quality,direction] = distCQM(X,d,estLabels,tau)
% Subspace-based clustering quality metrics (CQMs)
%   [quality,direction] = distCQM(X,d,estLabels)
% Input:
%   X: data array of size ambient dimension x number of data points
%   d: dimension of subspaces (assumed all equal)
%   estLabels: estimated labels
% Output:
%   quality: array of CQM values (more than described in paper)
%       [DB; DBS; DI; DI31; DI41; DI51; DI33; DI43; DI53; SI; CH; 
%           CS; COP; modCH]
%   direction: +1 if larger is better, -1 if smaller is better
%--------------------------------------------------------------------------
% Copyright @ John Lipor, 2020
%--------------------------------------------------------------------------

[D,N] = size(X);
X = X./repmat(sqrt(sum(X.^2,1)),D,1);
X(isnan(X)) = 0;
K = max(estLabels);

% compute distances from points to subspaces and std vectors
fullResids = zeros(N,K);
fullU = cell(K,1);
fullNk = zeros(K,1);
for kk = 1:K
    Xk = X(:,estLabels==kk);
    [Dk,Nk] = size(Xk);
    fullNk(kk) = Nk;
    if Dk < Nk
        [U,V] = eig(Xk*Xk');
        [v,vind] = sort(diag(V),'descend');
        Uk = U(:,vind(1:d));
        fullU{kk} = Uk;
    else
        [U,S,~] = svd(Xk);
        Uk = U(:,1:d);
        fullU{kk} = Uk;
    end
    fullResids(:,kk) = norms(X - Uk*(Uk\X));
end

% subspace-based pairwise distances
state.Ufull = fullU;
state.fullResids = fullResids;
Dmat = ssPdist(X,d,estLabels,state);

% mean distance of points to their respective clusters
meanSSDists = zeros(K,1);
for kk = 1:K
    meanSSDists(kk) = mean(fullResids(estLabels==kk,kk));
end

% distance matrix of subspaces
ssDmat = zeros(K);
for jj = 1:K
    Uj = fullU{jj};
    for kk = jj+1:K
        Uk = fullU{kk};       
        ssDmat(jj,kk) = ssDist(Uj,Uk);
    end
end
ssDmat = ssDmat + ssDmat';
ssDmat = ssDmat + diag(diag(inf*eye(K)));   % set diagonal to inf for searches later

%% Davies-Bouldin (lower better)
allInds = zeros(K);
for jj = 1:K
    for kk = jj+1:K
        allInds(jj,kk) = (meanSSDists(jj) + meanSSDists(kk))/ssDmat(jj,kk);
    end
end
allInds = allInds + allInds';
DB = mean(max(allInds,[],1));

%% Davies-Bouldin* (lower better)
coh = zeros(K,1);
sep = zeros(K,1);
for kk = 1:K
    sumMeanDists = meanSSDists(kk) + meanSSDists(setdiff(1:K,kk));
    coh(kk) = max(sumMeanDists);
    sep(kk) = min(ssDmat(kk,:),[],2);
end
DBS = sum(coh./sep)/K;

%% Dunn (higher better)
% smallest inter-cluster pdist and largest intra-cluster pdist 
minInterDists = zeros(K);
maxIntraDists = zeros(K,1);
d3 = zeros(K);
for jj = 1:K
    jInds = find(estLabels==jj);
    jDists = Dmat(jInds,jInds);
    maxIntraDists(jj) = max(jDists(:));
    for kk = jj+1:K
        kInds = find(estLabels==kk);
        jkDists = Dmat(jInds,kInds);
        minInterDists(jj,kk) = min(jkDists(:));
        d3(jj,kk) = sum(jkDists(:))/(length(jInds)*length(kInds));
        d5(jj,kk) = (sum(fullResids(jInds,jj)) + sum(fullResids(kInds,kk)))/(length(jInds) + length(kInds));
    end
end
minInterDists = minInterDists + minInterDists';

dmin = min(minInterDists(minInterDists~=0));
dmax = max(maxIntraDists);
DI = dmin/dmax;

%% Generalized Dunn (higher better)
D3 = 2*meanSSDists;
DI31 = min(d3(d3~=0))/dmax;
DI41 = min(ssDmat(:))/dmax;
DI51 = min(d5(d5~=0))/dmax;
DI33 = min(d3(d3~=0))/max(D3);
DI43 = min(ssDmat(:))/max(D3);
DI53 = min(d5(d5~=0))/max(D3);

%% Silhouette (higher better)
% a = mean intra-cluster pdist
% b = min inter-cluster pdist
meanDists = zeros(N,K);
for jj = 1:K
    jInds = find(estLabels==jj);
    for kk = 1:K
        kInds = find(estLabels==kk);
        if jj == kk
            meanDists(jInds,kk) = sum(Dmat(jInds,kInds),2)/(length(kInds)-1);
        else
            meanDists(jInds,kk) = sum(Dmat(jInds,kInds),2)/length(kInds);
        end
    end
end
% mean intra-cluster distance for each point
estInds = sub2ind([N,K],1:N,estLabels');
a = meanDists(estInds)';
% min inter-cluster distance for each point
bDists = meanDists;
bDists(estInds) = inf;
b = min(bDists,[],2);
% silhouette widths
s = (b-a)./max([a b],[],2);
% mean silhouette width for each cluster
sk = zeros(K,1);
for kk = 1:K
    kInds = find(estLabels==kk);
    sk(kk) = sum(s(kInds))/length(kInds);
end
SI = sum(sk)/K;

%% Calinski-Harabasz (higher better)
[Ux,~,~] = svds(X,d);
coh = zeros(K,1);
sep = zeros(K,1);
for kk = 1:K
    coh(kk) = length(estLabels==kk)*ssDist(fullU{kk},Ux);
    sep(kk) = sum(fullResids(estLabels==kk,kk));
end
CH = (N-K)/(K-1)*sum(coh)/sum(sep);

%% Modified Calinski-Harabasz (higher better)
% meanIntraDists computed with DI above
sep = zeros(K,1);
for kk = 1:K
    sep(kk) = sum(unique(Dmat(estLabels==kk,estLabels==kk)));
end
modCH = (N-K)/(K-1)*(sum(coh)/sum(sep));

%% CS (lower better)
coh = zeros(K,1);
for kk = 1:K
    kLabels = find(estLabels==kk);
    Xk = X(:,kLabels);
    Nk = fullNk(kk);
    maxDists = max(Dmat(kLabels,kLabels),[],1);
    coh(kk) = sum(maxDists)/Nk;
end
sep = min(ssDmat);
CS = sum(coh)/sum(sep);

%% COP (lower better)
coh = Nk.*meanSSDists;
sep = zeros(K,1);
for kk = 1:K
    Dk = Dmat(estLabels==kk,estLabels~=kk);
    maxDists = max(Dk,[],1);
    sep(kk) = min(maxDists);
end
COP = sum(coh./sep)/N;

% outputs
quality = [DB; DBS; DI; DI31; DI41; DI51; DI33; DI43; DI53; SI; CH; CS; COP; modCH];
direction = [-1 -1 1 1 1 1 1 1 1 1 1 -1 -1 1];
