function [quality,direction] = subspaceCQM(X,d,estLabels)
% Subspace-based clustering quality metrics (CQMs)
%   [quality,direction] = subspaceCQM(X,d,estLabels)
% Input:
%   X: data array of size ambient dimension x number of data points
%   d: dimension of subspaces (assumed all equal)
%   estLabels: estimated labels
% Output:
%   quality: array of CQM values
%       [kssCost, RM, NRM, KC2]
%   direction: +1 if larger is better, -1 if smaller is better
%--------------------------------------------------------------------------
% Copyright @ John Lipor, 2020
%--------------------------------------------------------------------------

[D,N] = size(X);
X = X./repmat(sqrt(sum(X.^2,1)),D,1);
X(isnan(X)) = 0;
K = max(estLabels);

% compute distances from points to subspaces
allVals = zeros(N,K);
fullU = cell(K,1);
for kk = 1:K
    Xk = X(:,estLabels==kk);
    [Dk,Nk] = size(Xk);
    if Nk < d
        Uk = orth(Xk);
        fullU{kk} = Uk;
    elseif Dk < Nk
        [U,V] = eig(Xk*Xk');
        [v,vind] = sort(diag(V),'descend');
        Uk = U(:,vind(1:d));
        fullU{kk} = Uk;
    else
        [U,S,~] = svd(Xk);
        Uk = U(:,1:d);
        fullU{kk} = Uk;
    end
    allVals(:,kk) = norms(X - Uk*(Uk\X));
end

% compute margin of each point
[minVals,minInds] = sort(allVals,2,'ascend');
minInd = minInds(:,1:2);
ind1 = sub2ind([N,K],1:N,minInds(:,1)');
ind2 = sub2ind([N,K],1:N,minInds(:,2)');

% distance matrix of subspaces
ssDmat = zeros(K);
for jj = 1:K
    Uj = fullU{jj};
    for kk = jj+1:K
        Uk = fullU{kk};       
        ssDmat(jj,kk) = ssDist(Uj,Uk);
    end
end
ssDmat = ssDmat + ssDmat';
avgSubDist = mean(squareform(ssDmat));

% KSS cost
kssCost = sqrt(sum(allVals(ind1).^2))/norm(X,'fro');

% KSS cost with subspace distance penalization
normKSS = 0;
for kk = 1:K
    kInds = find(minInds(:,1)==kk);
    normKSS = normKSS + sum(minVals(kInds,1).^2) / min(ssDmat(kk,setdiff(1:K,kk)))^2;
end
normKSS = normKSS/N;

% outputs
quality = [kssCost; normKSS];
direction = [-1 -1];
